@@ -1,9 +1,13 @@
 #!/usr/bin/env python3
 # -*- coding: utf-8 -*-
 """
-Created on Thu Mar 21 10:46:40 2019
+Basic data structures that allow more efficient handling of year-country
+based data sets. 
+
+Datatables are based on pandas dataframes and enfore regions as index
+and years as columns. It uses meta data and provides full unit handling
+by any computations using datatables.
 
-@author: andreas geiges
 """
 
 import pandas as pd
@@ -13,13 +17,18 @@ from copy import copy
 
 from . import core
 from . import config 
-from . import mapping as mapp
+#from . import mapping as mapp
 from . import util
 #from . import io_tools
 #from .tools import xarray
             
 class Datatable(pd.DataFrame):
+    """
+    Datatable
+    ^^^^^^^^^
     
+    Datatable is derrived from pandas dataframe. 
+    """
     _metadata = ['meta', 'ID']
 
     def __init__(self, *args, **kwargs):
@@ -47,6 +56,24 @@ class Datatable(pd.DataFrame):
     
     @classmethod
     def from_pyam(cls, idf, **kwargs):
+        """
+        Create a datatable from an iam dataframe.
+
+        Parameters
+        ----------
+        cls : datatable class
+            DESCRIPTION.
+        idf : pyam dataframe
+            dataframe that contrains the data that is used to create the datatable.
+        **kwargs : TYPE
+            DESCRIPTION.
+
+        Returns
+        -------
+        datatatble : datatoolbox datatable
+            Datatable with original unit and related meta data.
+
+        """
         import pyam
 
         if kwargs:
@@ -81,6 +108,32 @@ class Datatable(pd.DataFrame):
 
         return cls(data, meta=meta)
 
+    @classmethod
+    def from_excel(cls, filepath, sheetName = None):
+        """
+        Create a dataframe from a suitable excel file that is saved by datatoolbox.
+        
+        Parameters
+        ----------
+        cls : class
+        
+        filepath : str
+            Path to the file.
+        sheetName : str, optional
+            Sheetn ame that is read in. The default is None.
+     
+        Returns
+        -------
+        datatable
+            DESCRIPTION.
+        """
+        if sheetName is None:
+            sheetNames = None
+        else:
+            sheetNames= [sheetNames]
+        return read_excel(filepath,
+                          sheetNames=sheetNames)
+       
     def _to_xarray(self):
         
         return core.xr.DataArray(self.values, coords=[self.index, self.columns], dims=['space','time'], attrs=self.meta)
@@ -109,7 +162,16 @@ class Datatable(pd.DataFrame):
     
    
     def __appendMetaData__(self, metaDict):    
-        
+        """
+        Private function to append meta data.
+
+        Parameters
+        ----------
+        metaDict : dict
+            New meta data.
+
+
+        """
         self.__setattr__('meta', metaDict.copy())
 
         #test if unit is recognized
@@ -119,14 +181,14 @@ class Datatable(pd.DataFrame):
 
     def copy(self, deep=True):
         """
-        Make a copy of this ClimateFrame object
+        Make a copy of this Datatable object
         Parameters
         ----------
         deep : boolean, default True
             Make a deep copy, i.e. also copy data
         Returns
         -------
-        copy : ClimateFrame
+        copy : Datatable
         """
         # FIXME: this will likely be unnecessary in pandas >= 0.13
         data = self._data
@@ -134,20 +196,53 @@ class Datatable(pd.DataFrame):
             data = data.copy(deep=True)
         return Datatable(data).__finalize__(self) 
 
-    def diff(self, periods=1, axis=0):
+    def diff(self, 
+             periods=1, 
+             axis=0):
+        """
+        Compute the difference between different years in the datatable
+        Equivalent do pandas diff but return datatable.
         
+        Parameters
+        ----------
+        periods : int, optional
+            DESCRIPTION. The default is 1.
+        axis : int, optional
+            DESCRIPTION. The default is 0.
+
+        Returns
+        -------
+        out : TYPE
+            DESCRIPTION.
+
+        """
         out = super(Datatable, self).diff(periods=periods, axis=axis)
         out.meta['unit'] = self.meta['unit']
         
         return out
     
-#with pd.ExcelWriter('the_file.xlsx', engine='openpyxl', mode='a') as writer: 
-#     data_filtered.to_excel(writer) 
-    x = pd.DataFrame()
-    x.diff()
-    
     def to_excel(self, fileName = None, sheetName = "Sheet0", writer = None, append=False):
-        if writer is None:
+        """
+        Save datatable to excel.
+
+        Parameters
+        ----------
+        fileName : str, optional
+            Relative file path. If None is provide, a writer is expected. The default is None.
+        sheetName : str, optional
+            Sheet name that is read in. The default is "Sheet0".
+        writer : pandas excel writer, optional
+            Pandas writer that is used instead opening a new one. The default is None.
+        append : bool, optional
+            If true, data is appended to the writer. The default is False.
+
+        Returns
+        -------
+        None.
+
+        """
+
+        if fileName is not None:
             if append:
                 writer = pd.ExcelWriter(fileName, 
                                         engine='openpyxl', mode='a',
@@ -164,13 +259,27 @@ class Datatable(pd.DataFrame):
         
         metaSeries.to_excel(writer, sheet_name=sheetName, header=None, columns=None)
         super(Datatable, self).to_excel(writer, sheet_name= sheetName, startrow=len(metaSeries))
-        writer.close()
+        
+        if fileName is not None:
+            writer.close()
         
         
                 
     
     def to_csv(self, fileName=None):
-        
+        """
+        Save the datatable to an annotated csv file.
+
+        Parameters
+        ----------
+        fileName : str, optional
+            Path to file. The default is None.
+
+        Returns
+        -------
+        None.
+
+        """
         if fileName is None:
             fileName = '|'.join([ self.meta[key] for key in config.ID_FIELDS]) + '.csv'
         else:
@@ -189,6 +298,25 @@ class Datatable(pd.DataFrame):
         fid.close()
 
     def to_pyam(self, **kwargs):
+        """
+        Conversion to pyam dataframe.
+
+        Parameters
+        ----------
+        **kwargs : TYPE
+            DESCRIPTION.
+
+        Raises
+        ------
+        AssertionError
+            DESCRIPTION.
+
+        Returns
+        -------
+        idf : pyam dataframe
+            DESCRIPTION.
+
+        """
         from pyam import IamDataFrame
 
         meta = {
@@ -213,9 +341,33 @@ class Datatable(pd.DataFrame):
                 idf.set_meta(meta[field], field)
  
         return idf
+       
+    def to_IamDataFrame(self, **kwargs):
+        """
+        Function to sustain backwars compatibility
+        Depreciated.
+
+        """
+        return self.to_pyam(**kwargs)
         
     def convert(self, newUnit, context=None):
-        
+        """
+        Convert datatable to different unit and returns converted
+        datatable.
+
+        Parameters
+        ----------
+        newUnit : str
+            New unit string in which the datatable should be converted.
+        context : str, optional
+            Optional context (e.g. GWPAR4). The default is None.
+
+        Returns
+        -------
+        datatable
+            Datatable converted in the new unit.
+
+        """
         if self.meta['unit'] == newUnit:
             return self
         
@@ -241,6 +393,25 @@ class Datatable(pd.DataFrame):
         return aggregate_region(self, mapping)
     
     def interpolate(self, method="linear", add_missing_years=False):
+        """
+        Interpoltate missing data between year with the option to add
+        missing years in the columns.
+
+        Parameters
+        ----------
+        method : sting, optional
+            Interpolation method. The default is "linear".
+            - linear
+        add_missing_years : bool, optional
+            If true, missing years within the time value range are added to 
+            the dataframe. The default is False.
+
+        Returns
+        -------
+        datatable
+            Interpoltated dataframe.
+
+        """
         from datatoolbox.tools.for_datatables import interpolate
         
         if add_missing_years:
@@ -251,10 +422,33 @@ class Datatable(pd.DataFrame):
         return interpolate(self, method)
     
     def clean(self):
+        """
+        Clean up the dataframe to only recogniszed regions, years and numeric values. 
+        Removed columns and rows with only nan values.
 
+        Returns
+        -------
+        datatable
+            DESCRIPTION.
+
+        """
         return util.cleanDataTable(self)
     
-    def filter(self, spaceIDs=None):
+    def filter(self, spaceIDs):
+        """
+        Filter dataframe based on a list of spatial IDs.
+
+        Parameters
+        ----------
+        spaceIDs : iterable of str
+            DESCRIPTION.
+
+        Returns
+        -------
+        datatable
+            DESCRIPTION.
+
+        """
         mask = self.index.isin(spaceIDs)
         return self.iloc[mask,:]
     
@@ -263,6 +457,13 @@ class Datatable(pd.DataFrame):
         """
         This methods returns the yearly change for all years (t1) that reported
         and and where the previous year (t0) is also reported
+        
+        Parameters
+        ----------
+        forward : bool 
+            If true, the yearly change is computed in the forward direction, otherwise
+            backwards.
+            Default is forward.
         """
 
         #%%
@@ -292,6 +493,15 @@ class Datatable(pd.DataFrame):
     
     #%%
     def generateTableID(self):
+        """
+        Generates the table ID based on the meta data of the table.
+
+        Returns
+        -------
+        datatable
+            DESCRIPTION.
+
+        """
         # update meta data required for the ID
         self.meta =  core._update_meta(self.meta)
         self.ID   =  core._createDatabaseID(self.meta)
@@ -301,10 +511,28 @@ class Datatable(pd.DataFrame):
 
     
     def source(self):
+        """
+        Return the source of the table 
+        """
         return self.meta['source']
 
     def append(self, other, **kwargs):
+        """
+        Append data to the datatable
+        
 
+        Parameters
+        ----------
+        other : datatable
+            New data that will be added to the datatable.
+        **kwargs : TYPE
+            Default pandas append arguments.
+
+        Returns
+        -------
+        datatable
+
+        """
         if isinstance(other,Datatable):
             
             if other.meta['entity'] != self.meta['entity']:
@@ -322,20 +550,25 @@ class Datatable(pd.DataFrame):
         # overwrite scenario
         out.meta['scenario'] = 'computed: ' + self.meta['scenario'] + '+' + other.meta['scenario']
         return out
-#    def append(self, other, **kwargs):
-#        if isinstance(other,Datatable):
-#            
-#            if other.meta['entity'] != self.meta['entity']:
-#                raise(BaseException('Physical entities do not match, please correct'))
-#            if other.meta['unit'] != self.meta['unit']:
-#                other = other.convert(self.meta['unit'])
-#            
-#            
-#        
-#        super(Datatable, self).append(self, other, **kwargs)
+
     
         
     def __add__(self, other):
+        """
+        Private function to add two dataframes. The added table is converted to
+        the unit of first table.
+
+        Parameters
+        ----------
+        other : datatble
+            Data to add.
+
+        Returns
+        -------
+        out : datatable
+            DESCRIPTION.
+
+        """
         if isinstance(other,Datatable):
             
             if self.meta['unit'] == other.meta['unit']:
@@ -349,8 +582,6 @@ class Datatable(pd.DataFrame):
             out.meta['unit'] = self.meta['unit']
             out.meta['source'] = 'calculation'
         else:
-#            import pdb
-#            pdb.set_trace()
             out = Datatable(super(Datatable, self).__add__(other))
             out.meta['unit'] = self.meta['unit']
             out.meta['source'] = 'calculation'
@@ -359,6 +590,21 @@ class Datatable(pd.DataFrame):
     __radd__ = __add__
     
     def __sub__(self, other):
+        """
+        Private function to subract two dataframes. The subracted table is converted to
+        the unit of first table.
+
+        Parameters
+        ----------
+        other : datatble
+            Data to substract.
+
+        Returns
+        -------
+        out : datatable
+            DESCRIPTION.
+
+        """
         if isinstance(other,Datatable):
             if self.meta['unit'] == other.meta['unit']:
                 factor = 1
@@ -375,6 +621,9 @@ class Datatable(pd.DataFrame):
         return out
     
     def __rsub__(self, other):
+        """
+        Equivalent to __sub__
+        """
         if isinstance(other,Datatable):
             if self.meta['unit'] == other.meta['unit']:
                 factor = 1
@@ -468,7 +717,25 @@ class Datatable(pd.DataFrame):
         return outStr
 #%%
 class TableSet(dict):
+    """
+    Class TableSet that is inherited from the dict class. It organized multiple 
+    heterogeneous datatbles into one structure.
+    """
     def __init__(self, IDList=None):
+        """
+        Create tableset from a given list of table IDs. All tables are loaded from
+        the database.
+
+        Parameters
+        ----------
+        IDList : list, optional
+            DESCRIPTION. The default is None.
+
+        Returns
+        -------
+        None.
+
+        """
         super(dict, self).__init__()
         self.inventory = pd.DataFrame(columns = ['key']+ config.INVENTORY_FIELDS)
         
@@ -477,52 +744,201 @@ class TableSet(dict):
                 self.add(core.DB.getTable(tableID))
         
         
-#        if config.AVAILABLE_XARRAY:
-#           self.to_Xarray = self._to_Xarray         
+    @classmethod
+    def from_list(cls,
+                  tableList):
+        """
+        Create tableSet form list of datatables.
+
+        Parameters
+        ----------
+        cls : TYPE
+            DESCRIPTION.
+        tableList : list
+            List of datatables.
+
+        Returns
+        -------
+        tableSet : tableset
+            DESCRIPTION.
+
+        """
+        tableSet = cls()
+        for table in tableList:
+            tableSet.add(table)
+    
+        return tableSet
+    
             
     def to_xarray(self, dimensions):
+        """
+        Convert tableset to and xarray with the given dimenions.
+        Requires xarray installed
+
+        Parameters
+        ----------
+        dimensions : list of str
+            List of xarray dimensions.
+
+        Returns
+        -------
+        xr.xarray
+            DESCRIPTION.
+
+        """
         if not config.AVAILABLE_XARRAY:
             raise(BaseException('module xarray not available'))
         return core.to_XDataArray(self, dimensions)
        
-    def to_xset(self):
+    def to_xset(self, dimensions = ['region', 'time']):
+        """
+        Convert table set to an xarray data set.
+
+        Parameters
+        ----------
+        dimensions : list, optional
+            DESCRIPTION. The default is ['region', 'time'].
+
+        Returns
+        -------
+        xr.Dataset
+            DESCRIPTION.
+
+        """
         dimensions = ['region', 'time']
         if not config.AVAILABLE_XARRAY:
             raise(BaseException('module xarray not available'))
         return core.to_XDataSet(self, dimensions)
     
+    def to_list(self):
+        """
+        Convert to list of tables
+
+        Returns
+        -------
+        list
+            List of datatables.
+
+        """
+        return [ self[key] for key in self.keys()]
     
     def __iter__(self):
         return iter(self.values())
     
     def add(self, datatables=None, tableID=None):
-        if isinstance(datatables, (list, TableSet)):
-            for datatable in datatables:
-                self._add(datatable, tableID)
-        else:
-            datatable = datatables
-            self._add(datatable, tableID)
+        """
+        Add new tables to table set. Either datatables or table IDs should be given.
+
+        Parameters
+        ----------
+        datatables : list of datatables, optional
+            DESCRIPTION. The default is None.
+        tableID : TYPE, optional
+            DESCRIPTION. The default is None.
+
+        Returns
+        -------
+        None.
+
+        """
+        
+        # assert only on parameter is None
+        assert not ((datatables is None) and (tableID is None))
+        
+        if datatables is not None:
             
-    def _add(self, datatable=None, tableID=None):
-        if datatable is None:
-            # adding only the ID, the full table is only loaded when necessary
-            self[tableID] = None
-            self.inventory.loc[tableID] = [None for x in config.ID_FIELDS]
-        else:
-            # loading the full table
-            if datatable.ID is None:
-                datatable.generateTableID()
-            self[datatable.ID] = datatable
-            self.inventory.loc[datatable.ID, "key"] = datatable.ID
-            self.inventory.loc[datatable.ID, config.INVENTORY_FIELDS] = [datatable.meta.get(x,None) for x in config.INVENTORY_FIELDS]
+            if isinstance(datatables, list):
+                self._add_list(datatables)
+                
+            elif isinstance(datatables, TableSet):
+                self._add_TableSet(datatables)
+                
+            elif isinstance(datatables, Datatable):
+                self._add_single_table(datatables)
+                
+            else:
+                 print('Data type not recognized.')
+            
+        elif tableID is not None:
+            self._add_tableID(tableID)
+            
+            
+            
+    def _add_list(self, tableList):
+        for table in tableList:
+            tableID = table.generateTableID()
+            
+            if tableID in self.keys():
+                self._update(table, tableID)
+            else:
+                self.__setitem__(tableID, table)
     
+    def _add_TableSet(self, tableSet):
+        
+        for tableID, table in tableSet.items():
+            
+            if tableID in self.keys():
+                self._update(table, tableID)
+            else:
+                self.__setitem__(tableID, table)
+    
+    def _add_single_table(self, table):
+        tableID = table.generateTableID()
+        if tableID in self.keys():
+            self._update(table, tableID)
+        else:
+            self.__setitem__(tableID, table)
+        
+    def _add_tableID(self, tableID):
+        self[tableID] = None
+        self.inventory.loc[tableID] = [None for x in config.ID_FIELDS]   
+        
+        
+    def _update(self, table, tableKey):
+        
+        # make sure the data is compatible
+#        print(table.meta)
+#        print(self[tableKey].meta)
+        assert table.meta == self[tableKey].meta
+        
+        # update data
+        self[tableKey] = pd.concat([self[tableKey] , table])
+            
+
     def remove(self, tableID):
+        """
+        Remove table form tableSet.
+
+        Parameters
+        ----------
+        tableID : str
+            TableID.
+
+        Returns
+        -------
+        None.
+
+        """
         del self[tableID]
         self.inventory.drop(tableID, inplace=True)
         
     
     def filter(self,**kwargs):
-        
+        """
+        Filter tableSet based on the given table inventory columns. 
+        (see config.INVENTORY_FIELDS)
+
+        Parameters
+        ----------
+        **kwargs : dict-like
+            Filter arguments as string that need to be contained in the fields..
+
+        Returns
+        -------
+        newTableSet : tableSet
+            DESCRIPTION.
+
+        """
         inv = self.inventory.copy()
         for key in kwargs.keys():
             #table = table.loc[self.inventory[key] == kwargs[key]]
@@ -536,18 +952,7 @@ class TableSet(dict):
             newTableSet[key] = self[key]
             
         return newTableSet
-#        def getInventory(self, **kwargs):
-#        
-#        table = self.inventory.copy()
-#        for key in kwargs.keys():
-#            #table = table.loc[self.inventory[key] == kwargs[key]]
-#            mask = self.inventory[key].str.contains(kwargs[key], regex=False)
-#            mask[pd.isna(mask)] = False
-#            mask = mask.astype(bool)
-#            table = table.loc[mask].copy()
-#            
-#        
-#        return table
+
     def aggregate_to_region(self, mapping):
         """ 
         This functions added the aggregates to the output according to the provided
@@ -575,15 +980,31 @@ class TableSet(dict):
             try:
                 datatable.generateTableID()
             except:
-                print('Cuold not generate ID, key used instead')
+#                print('Could not generate ID, key used instead')
                 datatable.ID = key
         self.inventory.loc[datatable.ID, "key"] = key
         self.inventory.loc[datatable.ID, config.INVENTORY_FIELDS] = [datatable.meta.get(x,None) for x in config.INVENTORY_FIELDS]
     
 
     
-    def to_excel(self, fileName, append=False):
-       
+    def to_excel(self, 
+                 fileName, 
+                 append=False):
+        """
+        Sace TableSet as excel file with individual datatables in individual sheets.
+    
+        Parameters
+        ----------
+        fileName : str
+            File path.
+        append : bool, optional
+            If true, try to append data. The default is False.
+    
+        Returns
+        -------
+        None.
+
+        """
 
         if append:
             writer = pd.ExcelWriter(fileName, 
@@ -674,7 +1095,7 @@ class TableSet(dict):
         # move id columns to the front
         id_cols = pd.Index(['variable', 'region', 'scenario', 'model', 'unit'])
         long_df = long_df[id_cols.union(long_df.columns)]
-        
+        long_df = pd.DataFrame(long_df)
         return long_df
 
     def to_pyam(self):
@@ -733,24 +1154,6 @@ class TableSet(dict):
         plt.xticks([x +.5 for x in range(len(avail.columns))], avail.columns, rotation=45)
 
 
-
-
-#table = dt.getTable(dt.inventory().index[0])
-#table2 = dt.getTable(dt.inventory().index[1])
-#tableset = TableSet()
-#tableset.add(table)
-#tableset['Table1'] = table
-#tableset['Table2'] = table2
-#    
-#print(tableset.pathways())
-#
-#cropsSet = tableset.filter(variable='Crop')
-# %%
-
-#%%    
-    
-    
-    
 class Visualization():
     """ 
     This class addes handy built-in visualizations to datatables
@@ -828,8 +1231,10 @@ class Visualization():
         
         source = ColumnDataSource(df)
         legend_it = list()
+        import datatoolbox as dt
         for spatID,color in zip(self.df.index, palette):
-            coName = mapp.countries.codes.name.loc[spatID]
+#            coName = mapp.countries.codes.name.loc[spatID]
+            coName = dt.mapp.nameOfCountry(spatID)
             #plot.line(x=self.columns, y=self.loc[spatID], source=source, name=spatID)
             c = plot.line('year', spatID + '_y', source=source, name=spatID, line_width=2, line_color = color)
             legend_it.append((coName, [c]))
@@ -930,8 +1335,9 @@ class Visualization():
         
         source = ColumnDataSource(df)
         legend_it = list()
+        import datatoolbox as  dt
         for spatID, color in zip(self.df.index, palette):
-            coName = mapp.countries.codes.name.loc[spatID]
+            coName = dt.mapp.countries.codes.name.loc[spatID]
             #plot.line(x=self.columns, y=self.loc[spatID], source=source, name=spatID)
             c = plot.circle('year', spatID + '_y', source=source, name=spatID, color = color)
             legend_it.append((coName, [c]))
@@ -1020,7 +1426,11 @@ def read_excel(fileName, sheetNames = None):
                                   index   = fileContent.loc[columnIdx+1:, 0], 
                                   columns = [int(x) for x in fileContent.loc[columnIdx, 1:]], 
                                   meta    = metaDict)
-            dataTable.generateTableID()
+            
+            try:
+                dataTable.generateTableID()
+            except:
+                print('Warning: Meta data incomplete, table ID not generated')
             out = dataTable
 #        except:
 #                print('Failed to read the sheet: {}'.format(sheet))
