@@ -7,8 +7,10 @@ Created on Thu May 28 19:56:03 2020
 """
 import copy
 import datatoolbox as dt
+
 from util import df, df2, sourceMeta
  
+dt.admin.switch_database_to_testing()
 
 def test_validate():
     dt.core.DB._validateRepository()
@@ -37,7 +39,7 @@ def test_update_meta():
     assert  'Emissions|CO2|transport__Historic__XYZ_2020' in dt.core.DB.inventory.index
     
 def test_delete_table():
-    dt.removeTable(df.ID)
+    dt.removeTable('Emissions|CO2|transport__Historic__XYZ_2020')
     
     assert not dt.isAvailable(df.ID)
   
@@ -66,6 +68,35 @@ def test_findp():
     
     assert len(inv.variable.unique()) ==2
     
+    
+def test_table_logging():
+    # create tempory table in DB
+    table = dt.getTable('Numbers|Ones__Historic__Numbers_2020')
+    tableNew = table*2
+    tableNew.meta.update(table.meta)
+    tableNew.meta.update({'category':'Twos'})
+    dt.commitTable(tableNew, 'test table')
+    def test_analysis():
+
+        table = dt.getTable('Numbers|Twos__Historic__Numbers_2020')
+        print(table.sum())
+
+    # run to save all required table for the analysis
+    dt.core.DB.startLogTables()
+    test_analysis()
+    dt.core.DB.stopLogTables()
+    dt.core.DB.save_logged_tables()
+    
+    # remove table form database
+    dt.removeTable('Numbers|Twos__Historic__Numbers_2020')
+    
+    #run analysis with missing table, but locally stored
+    test_analysis()    
+    
+    #cleanup 
+    import shutil
+    shutil.rmtree('data')
+    
 if __name__ == '__main__':
     test_commit_new_table()
     test_validate_ID()
@@ -75,3 +106,4 @@ if __name__ == '__main__':
     test_commit_mutliple_tables()
     test_delete_mutliple_tables()
     test_delete_source()
+    test_table_logging()
\ No newline at end of file
